from functools import wraps
from typing import Any, Union
import typing

from aas_middleware.connect.connectors.connector import Connector, Consumer, Provider
from aas_middleware.connect.workflows.workflow import Workflow
from aas_middleware.middleware.registries import ConnectionInfo, PersistenceConnectionRegistry
from aas_middleware.model.data_model import DataModel
from aas_middleware.model.formatting.formatter import Formatter
from aas_middleware.model.mapping.mapper import Mapper
from aas_middleware.model.util import get_value_attributes


async def update_persistence_with_value(persistence_connector: Connector, connection_info: ConnectionInfo, value: Any):
    if not connection_info.contained_model_id and not connection_info.field_id:
        # TODO: handle case when id of model changes...
        # TODO: also handle case when new models are added in the data model
        await persistence_connector.consume(value)
    elif not connection_info.field_id:
        persistence_model = await persistence_connector.provide()
        persistence_model_data_model = DataModel.from_models(persistence_model)
        # TODO: make sure when id changes the connection info changes as well
        persistence_contained_model = persistence_model_data_model.get_model(connection_info.contained_model_id)
        referencing_models = persistence_model_data_model.get_referencing_models(persistence_contained_model)
        for referencing_model in referencing_models:
            for attribute_name, attribute_value in get_value_attributes(referencing_model):
                if attribute_value == persistence_contained_model:
                    setattr(referencing_model, attribute_name, value)
        await persistence_connector.consume(persistence_model)
    elif not connection_info.contained_model_id:
        persistence_model = await persistence_connector.provide()
        setattr(persistence_model, connection_info.field_id, value)
        await persistence_connector.consume(persistence_model)
    else:
        persistence_model = await persistence_connector.provide()
        persistence_model_data_model = DataModel.from_models(persistence_model)
        persistence_contained_model = persistence_model_data_model.get_model(connection_info.contained_model_id)
        setattr(persistence_contained_model, connection_info.field_id, value)
        await persistence_connector.consume(persistence_model)


async def get_persistence_value(persistence_connector: Connector, connection_info: ConnectionInfo) -> Any:
    persistence_model = await persistence_connector.provide()
    if not connection_info.contained_model_id and not connection_info.field_id:
        return persistence_model
    elif not connection_info.field_id:
        persistence_model_data_model = DataModel.from_models(persistence_model)
        persistence_contained_model = persistence_model_data_model.get_model(connection_info.contained_model_id)
        return persistence_contained_model
    elif not connection_info.contained_model_id:
        return getattr(persistence_model, connection_info.field_id)
    else:
        persistence_model_data_model = DataModel.from_models(persistence_model)
        persistence_contained_model = persistence_model_data_model.get_model(connection_info.contained_model_id)
        return getattr(persistence_contained_model, connection_info.field_id)
    


def adjust_body_for_persistence_schema(body: Any, external_mapper: typing.Optional[Mapper] = None, formatter: typing.Optional[Formatter] = None) -> Any:
    """
    Modifies the body for persistence.

    Args:
        body (Any): The body to modify.
        external_mapper (typing.Optional[Mapper], optional): The mapper that maps the body to the persistence model. Defaults to None.
        formatter (typing.Optional[Formatter], optional): The formatter that serializes the body. Defaults to None.

    Returns:
        Any: The modified body.
    """
    if formatter:
        body = formatter.deserialize(body)
    if external_mapper:
        body = external_mapper.map(body)
    return body


def adjust_body_for_external_schema(body: Any, persistence_mapper: typing.Optional[Mapper] = None, formatter: typing.Optional[Formatter] = None) -> Any:
    """
    Modifies the body for an external schema.

    Args:
        body (Any): The body to modify.
        persistence_mapper (typing.Optional[Mapper], optional): The mapper that maps the body to the external model. Defaults to None.
        formatter (typing.Optional[Formatter], optional): The formatter that serializes the body. Defaults to None.

    Returns:
        Any: The modified body.
    """
    if persistence_mapper:
        body = persistence_mapper.map(body)
    if formatter:
        body = formatter.serialize(body)
    return body


def synchronize_connector_with_persistence(connector: Union[Consumer, Provider], connection_info: ConnectionInfo, persistence_registry: PersistenceConnectionRegistry, persistence_mapper: typing.Optional[Mapper] = None, external_mapper: typing.Optional[Mapper] = None, formatter: typing.Optional[Formatter] = None):
    """
    Synchronizes a connector with the persistence layer.

    Args:
        connector (Union[Consumer, Provider]): The connector to synchronize.
        connection_info (ConnectionInfo): The connection info for the persistence layer.
        persistence_registry (PersistenceConnectionRegistry): The registry for the persistence connectors.    
    """
    if isinstance(connector, Consumer):
        original_consume = connector.consume

        @wraps(connector.consume)
        async def wrapped_consume(consumer_body: Any):
            persistence_connector = persistence_registry.get_connector_by_data_model_and_model_id(data_model_name=connection_info.data_model_name, model_id=connection_info.model_id)
            if consumer_body is None:
                # TODO: make data model connection possible here, not possible because of some conventions in the registries
                persistence_body = await get_persistence_value(persistence_connector, connection_info)
                consumer_body = adjust_body_for_external_schema(persistence_body, persistence_mapper, formatter)
            else:
                persistence_body = adjust_body_for_persistence_schema(consumer_body, external_mapper, formatter)
                await update_persistence_with_value(persistence_connector, connection_info, persistence_body)
            await original_consume(consumer_body)

        connector.consume = wrapped_consume

    if isinstance(connector, Provider):
        original_provide = connector.provide

        @wraps(connector.provide)
        async def wrapped_provide() -> Any:
            persistence_connector = persistence_registry.get_connector_by_data_model_and_model_id(data_model_name=connection_info.data_model_name, model_id=connection_info.model_id)
            provider_body = await original_provide()
            persistence_body = await get_persistence_value(persistence_connector, connection_info)
            provider_body_in_peristence_schema = adjust_body_for_persistence_schema(provider_body, external_mapper, formatter)
            if provider_body_in_peristence_schema != persistence_body:
                await update_persistence_with_value(persistence_connector, connection_info, provider_body_in_peristence_schema)
                assert provider_body_in_peristence_schema == await get_persistence_value(persistence_connector, connection_info), f"Persistence value was not updated correctly. Expected {provider_body_in_peristence_schema}, got {await get_persistence_value(persistence_connector, connection_info)}"
            return provider_body

        connector.provide = wrapped_provide


def synchronize_workflow_with_persistence_consumer(workflow: Workflow, connection_info: ConnectionInfo, persistence_registry: PersistenceConnectionRegistry, external_mapper: typing.Optional[Mapper] = None, formatter: typing.Optional[Formatter] = None):
    """
    Synchronizes a workflow with a persistence consumer.

    Args:
        workflow (Workflow): The workflow to synchronize.
        arg_name (str): The name of the argument in the workflow that is a consumer.
        connection_info (ConnectionInfo): The connection info for the persistence layer.
        persistence_registry (PersistenceConnectionRegistry): The registry for the persistence connectors.
    """
    original_execute = workflow.execute

    @wraps(workflow.execute)
    async def wrapped_execute(execute_body: Any):
        print("execute body: ", execute_body)
        workflow_return = await original_execute(execute_body)
        print("workflow return: ", workflow_return)
        persistence_connector = persistence_registry.get_connector_by_data_model_and_model_id(data_model_name=connection_info.data_model_name, model_id=connection_info.model_id)
        persistence_body = adjust_body_for_persistence_schema(workflow_return, external_mapper, formatter)
        print("persistence body: ", persistence_body)
        await update_persistence_with_value(persistence_connector, connection_info, persistence_body)
        return workflow_return
    
    workflow.execute = wrapped_execute

def synchronize_workflow_with_persistence_provider(workflow: Workflow, arg_name: str, connection_info: ConnectionInfo, persistence_registry: PersistenceConnectionRegistry, persistence_mapper: typing.Optional[Mapper] = None, external_mapper: typing.Optional[Mapper]=None, formatter: typing.Optional[Formatter] = None):
    """
    Synchronizes a workflow with a persistence provider.

    Args:
        workflow (Workflow): _description_
        arg_name (str): _description_
        connection_info (ConnectionInfo): _description_
        persistence_registry (PersistenceConnectionRegistry): _description_
        persistence_mapper (typing.Optional[Mapper], optional): _description_. Defaults to None.
        external_mapper (typing.Optional[Mapper], optional): _description_. Defaults
        formatter (typing.Optional[Formatter], optional): _description_. Defaults to None.
    """
    original_execute = workflow.execute

    @wraps(workflow.execute)
    async def wrapped_execute(execute_body: Any):
        persistence_connector = persistence_registry.get_connector_by_data_model_and_model_id(data_model_name=connection_info.data_model_name, model_id=connection_info.model_id)

        if not execute_body:
            persistence_body = await get_persistence_value(persistence_connector, connection_info)
            execute_body = adjust_body_for_external_schema(persistence_body, persistence_mapper, formatter)
        else:
            persistence_body = adjust_body_for_persistence_schema(execute_body, external_mapper, formatter)
            await update_persistence_with_value(persistence_connector, connection_info, persistence_body)

        workflow_return = await original_execute(execute_body)
        return workflow_return
    
    workflow.execute = wrapped_execute




from __future__ import annotations

import json
from typing import Dict, List, Set, Tuple, TypeVar, Union, Any, Type
from datetime import datetime

from pydantic import BaseModel, ValidationError

from aas_middleware.model.core import Identifiable

from aas_middleware.model.reference_finder import ReferenceFinder, ReferenceInfo, patch_references
from aas_middleware.model.util import (
    convert_under_score_to_camel_case_str,
    convert_camel_case_to_underscrore_str,
    get_id_with_patch,
    get_value_attributes,
    is_identifiable_container,
    models_are_equal,
)

from aas_middleware.model.util import (
    replace_attribute_with_model,
)


NESTED_DICT = Dict[str, Union[Any, "NESTED_DICT"]]
T = TypeVar("T")


class DateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        return super().default(o)


class DataModel(BaseModel):
    """
    The data model is a container that allows to store all models of a data model and provides methods to access them easily by their id or type.

    Args:
        **data (Dict[str, Any]): The data to load into the data model (used by pydantic).

    Attributes:
        _models_key_id (Dict[str, Identifiable]): The dictionary of models with their id as key.
        _top_level_models (Dict[str, List[str]]): The dictionary of top level models with their type as key.
        _models_key_type (Dict[str, List[str]]): The dictionary of models with their type as key.
        _reference_info_dict_for_referencing (Dict[str, Dict[str, ReferenceInfo]]): The dictionary of reference infos with keys from the referencing model to the referenced model.
        _reference_info_dict_for_referenced (Dict[str, Dict[str, ReferenceInfo]]): The dictionary of reference infos with keys from the referenced model to the referencing model.
    """

    _key_ids_models: Dict[str, Identifiable] = {}
    _top_level_models: Dict[str, List[str]] = {}
    _models_key_type: Dict[str, List[str]] = {}
    _reference_infos: Set[ReferenceInfo] = set()
    _reference_info_dict_for_referencing: Dict[str, Dict[str, ReferenceInfo]] = {}
    _reference_info_dict_for_referenced: Dict[str, Dict[str, ReferenceInfo]] = {}

    _schemas: Dict[str, Type[Any]] = {}
    _top_level_schemas: Set[str] = set()
    _schema_reference_infos: Set[ReferenceInfo] = set()
    _schema_reference_info_for_referencing: Dict[str, Dict[str, ReferenceInfo]] = {}
    _schema_reference_info_for_referenced: Dict[str, Dict[str, ReferenceInfo]] = {}

    def __init__(self, **data: Dict[str, Any]):
        super().__init__(**data)
        try:
            Identifiable.model_validate(self)
            self.add_model(self)
        except ValidationError:
            for attribute_value in get_value_attributes(self).values():
                if is_identifiable_container(attribute_value):
                    self.add(*attribute_value)
                else:
                    self.add(attribute_value)

    @classmethod
    def from_models(
        cls, *models: Tuple[Identifiable], **data: Dict[str, Any]
    ) -> DataModel:
        """
        Method to create a data model from a list of provided models.

        Args:
            models (Tuple[Identifiable]): The models to load into the data model.
            data (Dict[str, Any]): The data to load into the data model.

        Returns:
            DataModel: The data model with loaded models
        """
        data_model = cls(**data)
        data_model.add(*models)
        return data_model
    
    @classmethod
    def from_model_types(cls, *model_types: Tuple[Type[Identifiable]], **data: Dict[str, Any]) -> DataModel:
        """
        Method to create a data model a provided list of model types.

        Args:
            model_types (Tuple[Type[Identifiable]]): The model types to load into the data model.
            data (Dict[str, Any]): The data to load into the data model.

        Returns:
            DataModel: The data model with loaded models
        """
        data_model = cls(**data)
        for model_type in model_types:
            data_model.add_schema(model_type)
        return data_model

    @property
    def model_ids(self) -> Set[str]:
        """
        Property to get the ids of all contained models.

        Returns:
            Set[str]: The set of ids.
        """
        return set(str(key) for key in self._key_ids_models.keys())

    def get_contained_ids(self) -> Set[str]:
        """
        Method to get all ids of contained models.

        Returns:
            Set[str]: The set of ids.
        """
        return self.model_ids

    def add(self, *models: Identifiable) -> None:
        """
        Method to add models to the data model.

        Args:
            *models (Tuple[Identifiable]): The models to load into the data model.
        """
        for model in models:
            self.add_model(model)

    def add_schema(self, schema: Type[Identifiable]) -> None:
        """
        Method to add a schema of the data model.

        Args:
            schema (Type[Identifiable]): The schema to load.
        """
        all_schemas, schema_reference_infos = ReferenceFinder.find_schema_references(schema)
        self._add_contained_schemas(all_schemas)
        self._add_top_level_schema(schema)
        self._add_schema_references_to_referencing_schemas_dict(schema_reference_infos)

    def add_model(self, model: Identifiable) -> None:
        """
        Method to load a model of the data model.

        Args:
            model (Identifiable): The model to load.
        """
        # Identifiable.model_validate(model)
        model_id = get_id_with_patch(model)
        if model_id in self.model_ids:
            raise ValueError(f"Model with id {model_id} already loaded.")
        self.add_schema(type(model))
        all_identifiables, reference_infos = ReferenceFinder.find(model)
        self._add_contained_models(model, all_identifiables)
        self._add_top_level_model(model)
        self._add_references_to_referencing_models_dict(reference_infos)

    def check_different_model_with_same_id_contained(self, model: Identifiable) -> bool:
        """
        Method to check if a model is already contained in the data model.

        Args:
            model (Identifiable): The model to check.

        Returns:
            bool: True if the model is already contained, False otherwise.
        """
        model_id = get_id_with_patch(model)
        if model_id in self.model_ids:
            same_id_model = self.get_model(model_id)
            if not same_id_model == model:
                return True
        return False

    def _add_contained_models(
        self, top_level_model: Identifiable, contained_models: List[Identifiable]
    ) -> None:
        """
        Method to load all contained models of a model.

        Args:
            top_level_model (Identifiable): The top level model to load.
            contained_models (List[Identifiable]): The contained models to load.
        """
        for contained_model in contained_models:
            contained_model_id = get_id_with_patch(contained_model)
            if contained_model_id in self.model_ids:
                same_id_model = self.get_model(contained_model_id)
                if not models_are_equal(same_id_model, contained_model):
                    raise ValueError(
                        f"Model with id {contained_model_id} already loaded but with different content. Make sure to only load models with unique ids."
                    )
                replace_attribute_with_model(top_level_model, same_id_model)
                continue
            self._add_model(contained_model)

    def _add_contained_schemas(
        self, contained_schemas: List[Type[Identifiable]]
    ) -> None:
        """
        Method to load all contained schemas of a schema.

        Args:
            top_level_schema (Type[Identifiable]): The top level schema to load.
            contained_schemas (List[Type[Identifiable]]): The contained schemas to load.
        """
        for contained_schema in contained_schemas:
            self._schemas[contained_schema.__name__] = contained_schema


    def patch_schema_references(self) -> None:
        """
        Function tries to patch schema reference infos to represent a more realistic data model.

        ReferenceInfos of Type Reference are patched so that another ReferenceInfo is added to every schema that has a class name that contains the reference id.
        ReferenceInfos of Type Attribute are patched so that another ReferenceInfo is added to every schema that has a class name that contains the reference id.
        """
        reference_infos = patch_references(self._schema_reference_infos, self._schemas.values())
        self._schema_reference_infos = reference_infos
        self._add_schema_references_to_referencing_schemas_dict(reference_infos)

    def _add_references_to_referencing_models_dict(
        self, reference_infos: Set[ReferenceInfo]
    ) -> None:
        """
        Method to add information about referencing model ids of the input model.

        Args:
            model (Identifiable): The model to add the information for.
        """
        self._reference_infos = self._reference_infos | reference_infos
        for reference_info in reference_infos:
            referencing_model_id = reference_info.identifiable_id
            referenced_model_id = reference_info.reference_id
            if not referencing_model_id in self._reference_info_dict_for_referencing:
                self._reference_info_dict_for_referencing[referencing_model_id] = {}
            self._reference_info_dict_for_referencing[referencing_model_id][
                referenced_model_id
            ] = reference_info
            if not referenced_model_id in self._reference_info_dict_for_referenced:
                self._reference_info_dict_for_referenced[referenced_model_id] = {}
            self._reference_info_dict_for_referenced[referenced_model_id][
                referencing_model_id
            ] = reference_info

    def _add_schema_references_to_referencing_schemas_dict(self, reference_infos: Set[ReferenceInfo]) -> None:
        """
        Method to add information about referencing schema ids of the input schema.

        Args:
            schema (Type[Identifiable]): The schema to add the information for.
        """
        self._schema_reference_infos = self._schema_reference_infos | reference_infos
        for reference_info in reference_infos:
            referencing_schema_id = reference_info.identifiable_id
            referenced_schema_id = reference_info.reference_id
            if not referencing_schema_id in self._schema_reference_info_for_referencing:
                self._schema_reference_info_for_referencing[referencing_schema_id] = {}
            self._schema_reference_info_for_referencing[referencing_schema_id][
                referenced_schema_id
            ] = reference_info
            if not referenced_schema_id in self._schema_reference_info_for_referenced:
                self._schema_reference_info_for_referenced[referenced_schema_id] = {}
            self._schema_reference_info_for_referenced[referenced_schema_id][
                referencing_schema_id
            ] = reference_info


    def _add_model(self, model: Identifiable) -> None:
        """
        Method to add a model to the data model.

        Args:
            model (Identifiable): The model to add.
        """
        model_id = get_id_with_patch(model)
        if model_id in self.model_ids:
            raise ValueError(f"Model with id {model_id} already loaded.")
        self._key_ids_models[model_id] = model
        type_name = model.__class__.__name__.split(".")[-1]
        if not type_name in self._models_key_type:
            self._models_key_type[type_name] = []
        self._models_key_type[type_name].append(model_id)

    def _add_top_level_model(self, model: Identifiable) -> None:
        """
        Method to add a model to the data model.

        Args:
            model (Identifiable): The model to add.
        """
        type_name = model.__class__.__name__.split(".")[-1]
        underscore_type_name = convert_camel_case_to_underscrore_str(type_name)
        if not underscore_type_name in self._top_level_models:
            self._top_level_models[underscore_type_name] = []
        self._top_level_models[underscore_type_name].append(get_id_with_patch(model))

    def _add_top_level_schema(self, schema: Type[Identifiable]) -> None:
        """
        Method to add a schema to the data model.

        Args:
            schema (Type[Identifiable]): The schema to add.
        """
        schema_name = schema.__name__
        if not schema_name in self._top_level_schemas:
            self._top_level_schemas.add(schema_name)

    def from_dict(self, data: NESTED_DICT, types: List[Type]) -> None:
        """
        Method to load a data model from a dict.

        Args:
            data (NESTED_DICT): The dict to load the data model from.
        """
        for attribute_name, attribute_value in data.items():
            class_name = convert_under_score_to_camel_case_str(attribute_name)
            for type_ in types:
                if type_.__name__ == class_name:
                    type_for_attribute_values = type_
                    break
            else:
                raise ValueError(f"Type {class_name} not supported.")
            for model_dict in attribute_value:
                model = type_for_attribute_values(**model_dict)
                self.add(model)

    def dict(self) -> NESTED_DICT:
        """
        Method to get the dict of the data model.

        Returns:
            NESTED_DICT: The dict of the data model.
        """
        nested_dict = {}
        for attribute_name, attribute_value in self.get_top_level_models().items():
            nested_dict[attribute_name] = [
                model.model_dump() for model in attribute_value
            ]
        return nested_dict

    def json(self) -> str:
        """
        Method to get the json of the data model.

        Returns:
            str: The json of the data model.
        """
        nested_dict = {}
        for attribute_name, attribute_value in self.get_top_level_models().items():
            # TODO: if a non-BaseModel object is loaded, this breakds down -> adjust this
            nested_dict[attribute_name] = [
                model.model_dump() for model in attribute_value
            ]
        return json.dumps(nested_dict, indent=4, cls=DateTimeEncoder)

    def get_top_level_models(self) -> Dict[str, List[Identifiable]]:
        """
        Method to get all models of the data model.

        Returns:
            Dict[str, List[Identifiable]]: The dictionary of models.
        """
        top_level_models = {}
        for top_level_model_name, top_level_model_ids in self._top_level_models.items():
            top_level_models[top_level_model_name] = [
                self.get_model(model_id) for model_id in top_level_model_ids
            ]
        return top_level_models
    
    def get_top_level_types(self) -> List[Type[Identifiable]]:
        """
        Method to get all types of the top level models in the data model.

        Returns:
            List[Type[Identifiable]]: The types of the top level models in the data model
        """
        top_level_types = []
        for schema_name in self._top_level_schemas:
            top_level_types.append(self._schemas[schema_name])
        return top_level_types

    def get_models_of_type_name(self, model_type_name: str) -> List[Identifiable]:
        """
        Method to get all models of a specific type.

        Args:
            model_type (str): The type of the models to get.

        Returns:
            List[Identifiable]: The list of models of the type.
        """
        if not model_type_name in self._models_key_type:
            raise ValueError(f"Model type {model_type_name} not supported.")
        return [
            self.get_model(model_id)
            for model_id in self._models_key_type[model_type_name]
        ]

    def get_models_of_type(self, model_type: Type[T]) -> List[T]:
        """
        Method to get all models of a specific type.

        Args:
            model_type (Type[T]): The type of the models to get.

        Returns:
            List[T]: The list of models of the type.
        """
        type_name = model_type.__name__.split(".")[-1]
        return self.get_models_of_type_name(type_name)

    def get_contained_models(self) -> List[Identifiable]:
        """
        Method to get all models that are contained in the data model.

        Returns:
            List[Identifiable]: The list of models.
        """
        return list(self._key_ids_models.values())

    def get_referencing_info(
        self, referenced_model: Identifiable
    ) -> List[ReferenceInfo]:
        """
        Method to get all reference infos of a model.

        Args:
            referenced_model (Identifiable): The model to get the reference infos for.

        Returns:
            List[ReferenceInfo]: The list of reference infos.
        """
        referenced_model_id = get_id_with_patch(referenced_model)
        if not referenced_model_id in self._reference_info_dict_for_referenced:
            return []
        return list(
            self._reference_info_dict_for_referenced[referenced_model_id].values()
        )

    def get_referencing_models(
        self, referenced_model: Identifiable
    ) -> List[Identifiable]:
        """
        Method to get all models that reference a specific model directly as an attribute or by its id.

        Args:
            referenced_model (Identifiable): The model to get the referencing models for.

        Returns:
            List[Identifiable]: The list of referencing models of the model.
        """
        referenced_model_id = get_id_with_patch(referenced_model)
        if not referenced_model_id in self._reference_info_dict_for_referenced:
            return []
        referencing_model_dict = self._reference_info_dict_for_referenced[
            referenced_model_id
        ]
        return [self.get_model(model_id) for model_id in referencing_model_dict]

    def get_referencing_models_of_type(
        self, referenced_model: Identifiable, referencing_model_type: Type[T]
    ) -> List[T]:
        """
        Method to get all models that reference a specific model directly as an attribute or by its id.

        Args:
            referenced_model (Identifiable): The model to get the referencing models for.
            referencing_model_type (Type[T]): The type of the referencing models to get.

        Returns:
            List[T]: The list of referencing models of the model.
        """
        referenced_model_id = get_id_with_patch(referenced_model)
        if not referenced_model_id in self._reference_info_dict_for_referenced:
            return []
        referencing_model_dict = self._reference_info_dict_for_referenced[
            referenced_model_id
        ]
        return [
            self.get_model(model_id)
            for model_id in referencing_model_dict
            if isinstance(self.get_model(model_id), referencing_model_type)
        ]

    def get_referenced_info(
        self, referencing_model: Identifiable
    ) -> List[ReferenceInfo]:
        """
        Method to get all reference infos of a model.

        Args:
            referencing_model (Identifiable): The model to get the reference infos for.

        Returns:
            List[ReferenceInfo]: The list of reference infos.
        """
        referencing_model_id = get_id_with_patch(referencing_model)
        if not referencing_model_id in self._reference_info_dict_for_referencing:
            return []
        return list(
            self._reference_info_dict_for_referencing[referencing_model_id].values()
        )

    def get_referenced_models(
        self, referencing_model: Identifiable
    ) -> List[Identifiable]:
        """
        Method to get all models that are referenced by a specific model directly as an attribute or by its id.

        Args:
            referencing_model (Identifiable): The model to get the referenced models for.

        Returns:
            List[Identifiable]: The list of referenced models of the model.
        """
        referencing_model_id = get_id_with_patch(referencing_model)
        if not referencing_model_id in self._reference_info_dict_for_referencing:
            return []
        referenced_model_dict = self._reference_info_dict_for_referencing[
            referencing_model_id
        ]
        return [self.get_model(model_id) for model_id in referenced_model_dict if model_id in self.model_ids]

    def get_referenced_models_of_type(
        self, referencing_model: Identifiable, referenced_model_type: Type[T]
    ) -> List[T]:
        """
        Method to get all models that are referenced by a specific model directly as an attribute or by its id.

        Args:
            referencing_model (Identifiable): The model to get the referenced models for.
            referenced_model_type (Type[T]): The type of the referenced models to get.

        Returns:
            List[T]: The list of referenced models of the model.
        """
        referencing_model_id = get_id_with_patch(referencing_model)
        if not referencing_model_id in self._reference_info_dict_for_referencing:
            return []
        referenced_model_dict = self._reference_info_dict_for_referencing[
            referencing_model_id
        ]
        return [
            self.get_model(model_id)
            for model_id in referenced_model_dict
            if isinstance(self.get_model(model_id), referenced_model_type)
        ]

    def get_model(self, model_id: str) -> Identifiable:
        """
        Method to get a model by its id.

        Args:
            model_id (str): The id of the model to get.

        Returns:
            Identifiable: The model.
        """
        if model_id not in self.model_ids:
            return None
        return self._key_ids_models[model_id]

    def contains_model(self, model_id: str) -> bool:
        """
        Method to check if a model is contained in the data model.

        Args:
            model_id (str): The id of the model to check.

        Returns:
            bool: True if the model is contained, False otherwise.
        """
        if self.get_model(model_id) is not None:
            return True
        return False



from typing import Any, Literal, Optional, Type, TypeVar, Union, List
import typing
from pydantic import BaseModel, Field, create_model

from aas_middleware.model.data_model import DataModel
from aas_middleware.model.formatting.aas import aas_model
from aas_middleware.model.reference_finder import ReferenceType
from aas_middleware.model.util import (
    get_id_with_patch,
    get_value_attributes,
    is_identifiable,
    is_identifiable_container,
)


def get_below_patch_type(patch_type: type) -> type:
    if patch_type == aas_model.AAS:
        return aas_model.Submodel
    elif patch_type == aas_model.Submodel:
        return aas_model.SubmodelElementCollection
    else:
        return aas_model.SubmodelElementCollection


T = TypeVar(
    "T", bound=aas_model.AAS | aas_model.Submodel | aas_model.SubmodelElementCollection
)


def get_type_hints_for_attribute(
    model: Any, attribute_name: str
) -> Type:
    if isinstance(model, BaseModel):
        return model.model_fields[attribute_name].annotation
    else:
        return typing.get_type_hints(model.__init__)[attribute_name]

def get_patched_type(
    model: Any, attribute_name: str, attribute_value: Any
) -> Type[T]:
    # TODO: this function should work without the specified attribute_value...
    attribute_type_hints = get_type_hints_for_attribute(model, attribute_name)
    if typing.get_origin(attribute_type_hints) == Union	and type(None) in typing.get_args(attribute_type_hints):
        return Optional[type(attribute_value)]
    elif typing.get_origin(attribute_type_hints) == Literal:
        return attribute_type_hints
    elif typing.get_origin(attribute_type_hints) == list:
        return List[type(attribute_value[0])]
    return type(attribute_value)


def get_patched_aas_object(model: Any, patch_type: Type[T]) -> T:
    """
    Rebuilds an Identifiable object to an AAS object.

    Args:
        model (object): The object to rebuild.

    Returns:
        aas_model.AAS: The rebuilt AAS object.
    """
    model_id = get_id_with_patch(model)
    dict_dynamic_model_creation = {}
    dict_model_instantiation = {}
    dict_model_instantiation["id_short"] = model_id
    if not patch_type == aas_model.SubmodelElementCollection:
        dict_model_instantiation["id"] = model_id

    for attribute_name, attribute_value in get_value_attributes(model).items():
        below_patch_type = get_below_patch_type(patch_type)

        if is_identifiable(attribute_value):
            patched_attribute_value = get_patched_aas_object(
                attribute_value, patch_type=below_patch_type
            )
        elif is_identifiable_container(attribute_value):
            patched_attribute_value = []
            for element in attribute_value:
                patched_element = get_patched_aas_object(
                    element, patch_type=below_patch_type
                )
                patched_attribute_value.append(patched_element)
        else:
            patched_attribute_value = attribute_value

        attribute_patch_type = get_patched_type(model, attribute_name, patched_attribute_value)

        dict_dynamic_model_creation.update(
            {
                attribute_name: (
                    attribute_patch_type,
                    Field(examples=[patched_attribute_value]),
                )
            }
        )
        dict_model_instantiation.update({attribute_name: patched_attribute_value})
    new_model_type = create_model(
        model.__class__.__name__, **dict_dynamic_model_creation, __base__=patch_type
    )
    return new_model_type.model_validate(dict_model_instantiation)


class DataModelRebuilder:
    def __init__(self, data_model: DataModel):
        """
        Rebuilds a data model with either direct or indirect references.

        Args:
            data_model (DataModel): The data model to rebuild.
        """
        self.data_model = data_model

    def rebuild_data_model_with_associations(self) -> DataModel:
        """
        Rebuilds all models in the data model with assosiations.

        Returns:
            DataModel: The rebuilt data model.
        """
        raise NotImplementedError

    def rebuild_data_model_for_AAS_structure(self) -> DataModel:
        """
        Rebuilds the data model for AAS meta model structure by adjusting the associations and references and infering correct AAS types.

        Returns:
            DataModel: The rebuilt data model.
        """
        aas_candidates = []
        submodel_candidates = []
        top_level_models_list = []
        for models in self.data_model.get_top_level_models().values():
            top_level_models_list += models

        for model in top_level_models_list:
            if isinstance(model, aas_model.AAS):
                aas_candidates.append(model)
                continue
            if any(
                reference_info.reference_type == ReferenceType.ASSOCIATION
                for reference_info in self.data_model.get_referencing_info(model)
            ):
                continue
            if not all(
                is_identifiable(attribute_value)
                for attribute_value in get_value_attributes(model).values()
            ):
                continue
            aas_candidates.append(model)

        for model in top_level_models_list:
            if model in aas_candidates:
                continue
            submodel_candidates.append(model)

        submodel_objects = []
        for submodel_candidate in submodel_candidates:
            # TODO: remove the need to patch models that are already subclasses of submodel...
            patched_submodel_object = get_patched_aas_object(
                submodel_candidate, patch_type=aas_model.Submodel
            )
            submodel_objects.append(patched_submodel_object)

        aas_objects = []
        for aas_candidate in aas_candidates:
            # TODO: remove the need to patch models that are already subclasses of aas...
            patched_aas_object = get_patched_aas_object(
                aas_candidate, patch_type=aas_model.AAS
            )
            aas_objects.append(patched_aas_object)

        aas_data_model = DataModel.from_models(*aas_objects + submodel_objects)
        return aas_data_model

    def rebuild_data_model_with_references(self) -> DataModel:
        """
        Rebuilds all models in the data model with references.

        Returns:
            DataModel: The rebuilt data model.
        """
        raise NotImplementedError



# TODO: check, if the polling could be timed with an asyncio event
# -> then polling could be implemented with the same interface as event driven http_reqest_connector

import aiohttp
import anyio

from aas_middleware.connect.connectors.http_request_connector import HttpRequestConnector


class HttpPollingConnector(HttpRequestConnector):
    def __init__(self, url: str, interval: float = 1):
        self.url = url
        self.interval = interval
        self.previous_value: str = ""

    async def provide(self) -> str:
        while True:
            await anyio.sleep(self.interval)
            async with aiohttp.ClientSession() as session:
                async with session.get(self.url) as response:
                    response_test = await response.text()
                    if response_test != self.previous_value:
                        self.previous_value = response_test
                        return response_test



import typing

import logging

from aas_middleware.model.mapping.mapper import Mapper

logger = logging.getLogger(__name__)

from pydantic import BaseModel, ConfigDict

from aas_middleware.connect.connectors.connector import Connector, Consumer, Provider
from aas_middleware.connect.connectors.model_connector import ModelConnector
from aas_middleware.connect.workflows.worfklow_description import WorkflowDescription
from aas_middleware.connect.workflows.workflow import Workflow
from aas_middleware.middleware.persistence_factory import PersistenceFactory
from aas_middleware.model.core import Identifiable


class ConnectionInfo(BaseModel):
    """
    Class that contains the information of a connection of a provider and a consumer to the persistence layer.
    """
    data_model_name: str
    model_id: typing.Optional[str] = None
    contained_model_id: typing.Optional[str] = None
    field_id: typing.Optional[str] = None

    model_config = ConfigDict(frozen=True, protected_namespaces=())

    @property
    def connection_type(self) -> typing.Literal["data_model", "model", "contained_model", "field"]:
        if self.model_id:
            if self.contained_model_id:
                if self.field_id:
                    return "field"
                return "contained_model"
            return "model"
        return "data_model"
    

class ConnectionRegistry:
    """
    Class that manages the connections of the middleware.
    """

    def __init__(self):
        self.connectors: typing.Dict[str, Connector] = {}
        self.connection_types: typing.Dict[str, typing.Type[Connector]] = {}
        self.connections: typing.Dict[ConnectionInfo, typing.List[str]] = {}


    def get_connector_id(self, connector: Connector) -> str:
        """
        Function to get a connector id from the connection manager.

        Args:
            connector (Connector): The connector of the connection.

        Returns:
            str: The id of the connector.

        Raises:
            KeyError: If the connector is not in the connection manager.
        """
        for connector_id, connector_ in self.connectors.items():
            if connector == connector_:
                return connector_id
        raise KeyError(f"Connector {connector} is not in the connection manager.")


    def get_connector(self, connector_id: str) -> Connector:
        """
        Function to get a connector from the connection manager.

        Args:
            connector_id (str): The id of the connector.

        Returns:
            Connector: The connector of the connection.

        Raises:
            KeyError: If the connector id is not in the connection manager.
        """
        return self.connectors[connector_id]


    def add_connector(self, connector_id: str, connector: Connector, connection_type: typing.Type[Connector]):
        """
        Function to add a connector to the connection manager.

        Args:
            connector_id (str): The name of the connector.
            connector (Connector): The connector to be added.
            connection_type (typing.Type[Connector]): The type of the connector.
        """
        self.connectors[connector_id] = connector
        self.connection_types[connector_id] = connection_type

    def add_connection(self, connector_id: str, connection_info: ConnectionInfo, connector: Connector, type_connection_info: typing.Type[typing.Any]):
        """
        Function to add a connection to the connection manager.

        Args:
            connector_id (str): The id of the connector.
            connection_info (ConnectionInfo): The connection info of the connection.
            connector (Connector): The connector of the connection.
            type_connection_info (typing.Type[typing.Any]): The type of the connection info of the connection.
        """
        if not connection_info in self.connections:
            self.connections[connection_info] = []
        self.connections[connection_info].append(connector_id)
        self.add_connector(connector_id, connector, type_connection_info)

    def get_connections(self, connection_info: ConnectionInfo) -> typing.List[typing.Tuple[Connector, typing.Type[typing.Any]]]:
        """
        Function to get a connection from the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.

        Returns:
            Connector: The connector of the connection.
        """
        connector_ids = self.connections[connection_info]
        connections = []
        for connector_id in connector_ids:
            connections.append((self.get_connector(connector_id), self.connection_types[connector_id]))
        return connections
    
    def get_data_model_connection_info(self, data_model_name: str) -> typing.List[ConnectionInfo]:
        """
        Function to get the connection info of a data model.

        Args:
            data_model_name (str): The name of the data model.

        Returns:
            typing.Set[ConnectionInfo]: The connection info of the data model.
        """
        connection_infos = []
        for connection_info in self.connections:
            if not connection_info.data_model_name == data_model_name:
                continue
            connection_infos.append(connection_info)
        return connection_infos
    
    def get_model_connection_info(self, model_id: str) -> typing.List[ConnectionInfo]:
        """
        Function to get the connection info of a model.

        Args:
            model_id (str): The id of the model.

        Returns:
            typing.Set[ConnectionInfo]: The connection info of the model.
        """
        connection_infos = []
        for connection_info in self.connections:
            if not connection_info.model_id == model_id:
                continue
            connection_infos.append(connection_info)
        return connection_infos
    
    def get_field_connection_info(self, field_id: str) -> typing.List[ConnectionInfo]:
        """
        Function to get the connection info of a field.

        Args:
            field_id (str): The id of the field.

        Returns:
            typing.Set[ConnectionInfo]: The connection info of the field.
        """
        connection_infos = []
        for connection_info in self.connections:
            if not connection_info.field_id == field_id:
                continue
            connection_infos.append(connection_info)
        return connection_infos
    
    def get_type_connection_info(self, type_name: str) -> typing.List[ConnectionInfo]:
        """
        Function to get the connection info of a type.

        Args:
            type_name (str): The name of the type.

        Returns:
            typing.Set[ConnectionInfo]: The connection info of the type.
        """
        connection_infos = []
        for connection_info, connections in self.connections.items():
            for connection in connections:
                if not connection[1].__name__ == type_name:
                    continue
                connection_infos.append(connection_info)
        return connection_infos
    

class PersistenceConnectionRegistry(ConnectionRegistry):
    """
    Class that manages the connections of the middleware.
    """

    def __init__(self):
        super().__init__()
        self.connectors: typing.Dict[str, Connector] = {}
        self.connection_types: typing.Dict[str, typing.Type[Connector]] = {}
        self.connections: typing.Dict[ConnectionInfo, str] = {}
        self.persistence_factories: typing.Dict[ConnectionInfo, typing.List[typing.Tuple[PersistenceFactory, typing.Type[typing.Any]]]] = {}


    def get_connector_by_data_model_and_model_id(self, data_model_name: str, model_id: str) -> Connector:
        """
        Function to get a connector from the connection manager.

        Args:
            data_model_name (str): The name of the data model.
            model_id (str): The id of the model.

        Returns:
            Connector: The connector of the connection.

        Raises:
            KeyError: If the model id is not in the connection manager.
        """
        connector_id = data_model_name + model_id + "_persistence"
        return self.get_connector(connector_id)

    def add_persistence_factory(self, connection_info: ConnectionInfo, model_type: typing.Type[typing.Any], persistence_factory: PersistenceFactory):
        """
        Function to add a persistence factory to the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.
            persistence_factory (PersistenceFactory): The persistence factory of the connection.
        """
        if not connection_info in self.persistence_factories:
            self.persistence_factories[connection_info] = []
        self.persistence_factories[connection_info].append((model_type, persistence_factory))

    def get_default_persistence_factory(self, connection_info: ConnectionInfo, persisted_model_type: typing.Type[typing.Any]) -> PersistenceFactory:
        """
        Function to get the default persistence factory of a connection.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.

        Returns:
            PersistenceFactory: The default persistence factory of the connection.
        """
        data_model_connection_info = ConnectionInfo(data_model_name=connection_info.data_model_name)
        if not data_model_connection_info in self.persistence_factories:
            logger.warning(f"No persistence factory found for {data_model_connection_info}. Using default persistence factory.")
            return PersistenceFactory(ModelConnector)
        for model_type, persistence_factory in self.persistence_factories[data_model_connection_info]:
            if issubclass(persisted_model_type, model_type):
                return persistence_factory
        logger.warning(f"No persistence factory found for {data_model_connection_info} and model type {persisted_model_type.__name__}. Using default persistence factory.")
        return PersistenceFactory(ModelConnector)

    def add_to_persistence(self, connection_info: ConnectionInfo, model: Identifiable, persistence_factory: typing.Optional[PersistenceFactory]):
        """
        Function to add a persistent connection to the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.
        """
        if not persistence_factory:
            persistence_factory = self.get_default_persistence_factory(connection_info, type(model))
        connector = persistence_factory.create(model)
        self.add_connection(connection_info, connector, type(model))

    def add_connection(self, connection_info: ConnectionInfo, connector: Connector, type_connection_info: typing.Type[typing.Any]):
        """
        Function to add a connection to the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.
            connector (Connector): The connector of the connection.
        """
        connector_id = connection_info.data_model_name + connection_info.model_id + "_persistence"
        self.add_connector(connector_id, connector, type_connection_info)
        self.connections[connection_info] = connector_id


    def remove_connection(self, connection_info: ConnectionInfo):
        """
        Function to remove a connection from the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.
        """
        del self.connections[connection_info]
        # TODO: also delete connector and connection type

    def get_connection(self, connection_info: ConnectionInfo) -> Connector:
        """
        Function to get a connection from the connection manager.

        Args:
            connection_info (ConnectionInfo): The connection info of the connection.

        Returns:
            Connector: The connector of the connection.

        Raises:
            KeyError: If the connection info is not in the connection manager.
        """
        if connection_info in self.connections:
            return self.get_connector(self.connections[connection_info])
        raise KeyError(f"Data model Connection info {connection_info} is not in the connection manager.")

    def get_type_connection_info(self, type_name: str) -> typing.List[ConnectionInfo]:
        """
        Function to get the connection info of a type.

        Args:
            type_name (str): The name of the type.

        Returns:
            typing.Set[ConnectionInfo]: The connection info of the type.
        """
        connection_infos = []
        for connection_info, connector_id in self.connections.items():
            model_type = self.connection_types[connector_id]
            if not model_type.__name__ == type_name:
                continue
            connection_infos.append(connection_info)
        return connection_infos


class WorkflowRegistry:
    """
    Class that manages the workflows of the registry.
    """

    def __init__(self):
        self.workflows: typing.Dict[str, Workflow] = {}

        self.workflow_providers: typing.Dict[str, typing.List[typing.Tuple[ConnectionInfo, Provider]]] = {}
        self.workflow_consumers: typing.Dict[str, typing.List[typing.Tuple[ConnectionInfo, Consumer]]] = {}

    def add_workflow(self, workflow: Workflow):
        """
        Function to add a workflow to the registry.

        Args:
            workflow (Workflow): The workflow to be added.
        """
        self.workflows[workflow.get_name()] = workflow

    def add_provider_to_workflow(self, workflow_name: str, connection_info: ConnectionInfo, provider: Provider):
        """
        Function to add a provider to a workflow.

        Args:
            workflow_name (str): The name of the workflow.
            connection_info (ConnectionInfo): The connection info of the provider.
            provider (Provider): The provider to be added.
        """
        if not workflow_name in self.workflow_providers:
            self.workflow_providers[workflow_name] = []
        self.workflow_providers[workflow_name].append((connection_info, provider))

    def add_consumer_to_workflow(self, workflow_name: str, connection_info: ConnectionInfo, connector: Connector):
        """
        Function to add a consumer to a workflow.

        Args:
            workflow_name (str): The name of the workflow.
            connection_info (ConnectionInfo): The connection info of the consumer.
            connector (Connector): The connector to be added.
        """
        if not workflow_name in self.workflow_consumers:
            self.workflow_consumers[workflow_name] = []
        self.workflow_consumers[workflow_name].append((connection_info, connector))

    def get_workflows(self) -> typing.List[Workflow]:
        """
        Function to get the workflows in the registry.

        Returns:
            typing.List[Workflow]: The workflows in the registry.
        """
        return list(self.workflows.values())

    def get_workflow(self, workflow_name: str) -> Workflow:
        """
        Function to get a workflow from the registry.

        Args:
            workflow_name (str): The name of the workflow.

        Returns:
            Workflow: The workflow from the registry.

        Raises:
            KeyError: If the workflow is not in the registry.
        """
        return self.workflows[workflow_name]
    
    def get_connections_of_workflow(self, workflow_name: str) -> typing.Tuple[typing.List[typing.Tuple[ConnectionInfo, Provider]], typing.List[typing.Tuple[ConnectionInfo, Consumer]]]:
        """
        Function to get the connections of a workflow.

        Args:
            workflow_name (str): The name of the workflow.

        Returns:
            typing.Tuple[typing.List[ConnectionInfo, Provider], typing.List[ConnectionInfo, Consumer]]: The connections of the workflow.
        """
        return self.workflow_providers[workflow_name], self.workflow_consumers[workflow_name]

    
    def get_providers(self, workflow_name: str) -> typing.List[typing.Tuple[ConnectionInfo, Provider]]:
        """
        Function to get the providers of a workflow.

        Args:
            workflow_name (str): The name of the workflow.

        Returns:
            typing.List[typing.Tuple[ConnectionInfo, Provider]]: The providers of the workflow.
        """
        return self.workflow_providers[workflow_name]
    
    def get_consumers(self, workflow_name: str) -> typing.List[typing.Tuple[ConnectionInfo, Consumer]]:
        """
        Function to get the consumers of a workflow.

        Args:
            workflow_name (str): The name of the workflow.

        Returns:
            typing.List[typing.Tuple[ConnectionInfo, Consumer]]: The consumers of the workflow.
        """
        return self.workflow_consumers[workflow_name]
    
    def get_workflow_names(self) -> typing.List[str]:
        """
        Function to get the names of the workflows in the registry.

        Returns:
            typing.List[str]: The names of the workflows in the registry.
        """
        return list(self.workflows.keys())
    
    def get_workflow_descriptions(self) -> typing.List[WorkflowDescription]:
        """
        Function to get the descriptions of the workflows in the registry.

        Returns:
            typing.List[WorkflowDescription]: The descriptions of the workflows in the registry.
        """
        return [workflow.get_description() for workflow in self.workflows.values()]
    



class MapperRegistry:
    """
    Class that manages the mappers of the middleware.
    """

    def __init__(self):
        self.mappers: typing.Dict[str, Mapper] = {}
        self.mapper_input_types: typing.Dict[str, typing.Type] = {}
        self.mapper_output_types: typing.Dict[str, typing.Type] = {}

        self.connections: typing.Dict[typing.Tuple[ConnectionInfo, ConnectionInfo], str] = {}

    def add_mapper(self, mapper_id: str, mapper: Mapper, input_connection: typing.Optional[ConnectionInfo] = None, output_connection: typing.Optional[ConnectionInfo] = None):
        """
        Function to add a mapper to the registry.

        Args:
            mapper_id (str): The name of the mapper.
            mapper (Mapper): The mapper to be added.
            input_connection (typing.Optional[ConnectionInfo]): The input connection of the mapper.
            output_connection (typing.Optional[ConnectionInfo]): The output connection of the mapper.
        """
        self.mappers[mapper_id] = mapper
        self.mapper_input_types[mapper_id] = typing.get_type_hints(mapper.map)["data"]
        self.mapper_output_types[mapper_id] = typing.get_type_hints(mapper.map)["return"]
        if input_connection and output_connection:
            self.connections[(input_connection, output_connection)] = mapper_id
        elif input_connection or output_connection:
            raise ValueError("Either None or both input and output connection must be provided.")
                

    def get_mappers(self) -> typing.List[Mapper]:
        """
        Function to get the mappers in the registry.

        Returns:
            typing.List[Mapper]: The mappers in the registry
        """
        return list(self.mappers.values())

    def get_mapper(self, mapper_id: str) -> Mapper:
        """
        Function to get a mapper from the registry.

        Args:
            mapper_id (str): The id of the mapper.

        Returns:
            Mapper: The mapper from the registry.

        Raises:
            KeyError: If the mapper is not in the registry.
        """
        return self.mappers[mapper_id]
    
    def get_mapper_ids(self) -> typing.List[str]:
        """
        Function to get the names of the mappers in the registry.

        Returns:
            typing.List[str]: The names of the mappers in the registry.
        """
        return list(self.mappers.keys())
    
    def get_mapper_connections(self) -> typing.List[typing.Tuple[ConnectionInfo, ConnectionInfo]]:
        """
        Function to get the connections of the mappers in the registry.

        Returns:
            typing.List[typing.Tuple[ConnectionInfo, ConnectionInfo]]: The connections of the mappers in the registry.
        """
        return list(self.connections.keys())
    

    def get_mapper_by_input_connection(self, input_connection: ConnectionInfo) -> typing.Optional[Mapper]:
        """
        Function to get a mapper by the input connection.

        Args:
            input_connection (ConnectionInfo): The input connection of the mapper.

        Returns:
            Mapper: The mapper of the input connection.
        """
        for connection, mapper_id in self.connections.items():
            if connection[0] == input_connection:
                return self.mappers[mapper_id]
    
    def get_mapper_by_output_connection(self, output_connection: ConnectionInfo) -> typing.Optional[Mapper]:
        """
        Function to get a mapper by the output connection.

        Args:
            output_connection (ConnectionInfo): The output connection of the mapper.

        Returns:
            Mapper: The mapper of the output connection.
        """
        for connection, mapper_id in self.connections.items():
            if connection[1] == output_connection:
                return self.mappers[mapper_id]
            
    def get_connection_of_mapper(self, mapper_id: str) -> typing.List[typing.Tuple[ConnectionInfo, ConnectionInfo]]:
        """
        Function to get the connections of a mapper.

        Args:
            mapper_id (str): The id of the mapper.

        Returns:
            typing.List[typing.Tuple[ConnectionInfo, ConnectionInfo]]: The connections of the mapper.
        """
        connections = []
        for connection, mapper_id2 in self.connections.items():
            if mapper_id2 == self.mappers[mapper_id]:
                connections.append(connection)
        return connections
            

    
            



from __future__ import annotations
from enum import Enum
from typing import Annotated, TypeVar, Any
from uuid import UUID

from pydantic import BaseModel, BeforeValidator, model_validator


Identifier = TypeVar("Identifier", bound=str | int | UUID)
Reference = TypeVar("Reference", bound=str | int | UUID)
UnIdentifiable = (
    str | int | float | bool | None | UUID | Enum | list | tuple | set | type
)


def string_is_not_empty(v: str):
    assert v, "value must not be an empty string"
    return v


IdString = Annotated[str | int | UUID, BeforeValidator(string_is_not_empty)]


class Identifiable(BaseModel):
    """
    Base class for all identifiable classes that have an identifier, that allows to identify these objects.

    If no id is set, the id function of the python object is used. Otherwise, a uuid is generated.

    Args:
        id (str): Global id of the object.
    """

    id: IdString

    @model_validator(mode="before")
    @classmethod
    def check_id_and_id_short(cls, data: Any) -> Any:
        potential_id = get_id(data)
        assert potential_id, "Either id or id_short must be set"
        return {"id": potential_id}


from aas_middleware.model.util import get_id



from __future__ import annotations

from types import NoneType
from typing import List, Optional, Set, Tuple, Type, Union
import typing
from pydantic import BaseModel, ConfigDict
from enum import Enum

from aas_middleware.model.core import Identifiable
from aas_middleware.model.schema_util import get_all_contained_schemas, get_attribute_dict_of_schema
from aas_middleware.model.util import (
    get_all_contained_identifiables,
    get_id_with_patch,
    get_identifiable_types,
    get_reference_name,
    get_referenced_ids_of_model,
    get_identifiable_attributes_of_model,
    get_unidentifiable_attributes_of_model,
    is_identifiable_container,
    is_identifiable_type,
    is_identifiable_type_container,
)


class ReferenceType(Enum):
    """
    Enum for the reference types. There exist three types of references:
    - Association: The reference is done by an association between two objects, where the model has the referenced object as an attribute.
    - Reference: The reference element is an object and the reference in the model is done by referencing the id of the referenced object.
    - Attribute: The referenced element is an primitive attribute of the model.
    """
    ASSOCIATION = "association"
    REFERENCE = "reference"
    ATTRIBUTE = "attribute"


class ReferenceInfo(BaseModel):
    """
    Object reference to a model in the data model.

    Args:
        identifiable_id (str): The id of the identifiable.
        reference_id (str): The id of the referenced identifiable.
        reference_type (ReferenceType): The type of the reference.
    """

    identifiable_id: str
    reference_id: str
    reference_type: ReferenceType

    model_config = ConfigDict(frozen=True)


def get_reference_infos_of_model(model: Identifiable) -> Set[ReferenceInfo]:
    """
    Method to add information about referencing model ids of the input model.

    Args:
        model (Referable): The model to add the information for.

    Returns:
        Set[ReferenceInfo]: The list of reference infos.
    """
    reference_infos = set()
    identifiables_of_model = get_identifiable_attributes_of_model(model)
    for identifiable in identifiables_of_model:
        if identifiable == model:
            continue
        reference_info = ReferenceInfo(
            identifiable_id=get_id_with_patch(model),
            reference_id=get_id_with_patch(identifiable),
            reference_type=ReferenceType.ASSOCIATION,
        )
        reference_infos.add(reference_info)
    indirect_references = get_referenced_ids_of_model(model)
    for indirect_reference in indirect_references:
        reference_info = ReferenceInfo(
            identifiable_id=get_id_with_patch(model),
            reference_id=indirect_reference,
            reference_type=ReferenceType.REFERENCE,
        )
        reference_infos.add(reference_info)

    unidentifiable_attributes = get_unidentifiable_attributes_of_model(model)
    for attribute_name, attribute_value in unidentifiable_attributes.items():
        reference_info = ReferenceInfo(
            identifiable_id=get_id_with_patch(model),
            reference_id=f"{attribute_name}={attribute_value}",
            reference_type=ReferenceType.ATTRIBUTE,
        )
        reference_infos.add(reference_info)
    return reference_infos


def get_reference_infos(identifiables: List[Identifiable]) -> Set[ReferenceInfo]:
    """
    Method to get all reference infos of a list of identifiables.

    Args:
        identifiables (List[Identifiable]): The list of identifiables.

    Returns:
        Set[ReferenceInfo]: The list of reference infos.
    """
    reference_infos = set()
    for identifiable in identifiables:
        reference_infos = reference_infos | get_reference_infos_of_model(identifiable)
    return reference_infos


def get_reference_infos_of_schema(schema: Type[Identifiable]) -> Set[ReferenceInfo]:
    """
    Method to add information about referencing schema ids of the input schema.

    Args:
        schema (Type[Identifiable]): The schema to add the information for.

    Returns:
        Set[ReferenceInfo]: The list of reference infos.
    """
    reference_infos = set()
    attribute_dict_of_schema = get_attribute_dict_of_schema(schema)
    for attribute_name, attribute_type in attribute_dict_of_schema.items():
        attribute_types = get_identifiable_types(attribute_type)

        for arg in attribute_types:
            reference_info = get_reference_info_for_schema(schema, attribute_name, arg)
            if not reference_info:
                continue
            reference_infos.add(reference_info)
    return reference_infos


def get_reference_info_for_schema(schema: Type[Identifiable], attribute_name: str, attribute_type: Type[Identifiable]) -> Optional[ReferenceInfo]:
    if is_identifiable_type(attribute_type) or is_identifiable_type_container(attribute_type):
        return ReferenceInfo(
            identifiable_id=schema.__name__,
            reference_id=attribute_type.__name__,
            reference_type=ReferenceType.ASSOCIATION,
        )
    elif get_reference_name(attribute_name, attribute_type):
        return ReferenceInfo(
            identifiable_id=schema.__name__,
            reference_id=get_reference_name(attribute_name, attribute_type),
            reference_type=ReferenceType.REFERENCE,
        )
    else:
        return ReferenceInfo(
            identifiable_id=schema.__name__,
            reference_id=f"{schema.__name__}.{attribute_name}",
            reference_type=ReferenceType.ATTRIBUTE,
        )
    

def patch_references(references: Set[ReferenceInfo], schemas: List[Type[Identifiable]]) -> Set[ReferenceInfo]:
    patched_references = set()
    # FIXME: this is not working properly ---> make case distinctions and resolve them step by step
    schema_names = {schema.__name__.split(".")[-1] for schema in schemas}
    print(schema_names)
    for reference in references:
        if reference.reference_type == ReferenceType.ASSOCIATION:
            continue
        elif reference.reference_type == ReferenceType.ATTRIBUTE:
            for schema_name in schema_names:
                adjusted_reference_id = reference.reference_id.split(".")[-1]
                if adjusted_reference_id in schema_name and len(adjusted_reference_id) > len(schema_name) and not reference.identifiable_id == schema_name:
                    # print(f"Patch reference from {reference.identifiable_id} to {reference.reference_id} to {schema_name}")
                    patched_references.add(
                        ReferenceInfo(
                            identifiable_id=reference.reference_id,
                            reference_id=schema_name,
                            reference_type=ReferenceType.REFERENCE,
                        )
                    )
    
        elif reference.reference_type == ReferenceType.REFERENCE:
            for schema_name in schema_names:
                if reference.reference_id in schema_name and len(reference.reference_id) < len(schema_name) and not reference.identifiable_id == schema_name:
                    # print(f"Patch reference from {reference.identifiable_id} to {reference.reference_id} to {schema_name}")
                    patched_references.add(
                        ReferenceInfo(
                            identifiable_id=reference.reference_id,
                            reference_id=schema_name,
                            reference_type=ReferenceType.REFERENCE,
                        )
                    )

    return references | patched_references



def get_schema_reference_infos(schemas: List[Type[Identifiable]]) -> Set[ReferenceInfo]:
    """
    Method to get all reference infos of a list of schemas.

    Args:
        schemas (List[Type[Identifiable]]): The list of schemas.

    Returns:
        List[ReferenceInfo]: The list of reference infos.
    """
    reference_infos = set()
    for schema in schemas:
        reference_infos = reference_infos | get_reference_infos_of_schema(schema)
    return reference_infos



class ReferenceFinder:
    model: Identifiable
    contained_models: List[Identifiable] = []
    references: List[ReferenceInfo] = []

    contained_schemas: List[Type[Identifiable]] = []
    schema_references: List[ReferenceInfo] = []

    @classmethod
    def find(
        cls, model: Identifiable
    ) -> Tuple[List[Identifiable], Set[ReferenceInfo]]:
        """
        Method to find all contained models (inclusive the model itself) and references in a given model.

        Args:
            model (Identifiable): The model to find contained models and references in.

        Returns:
            Tuple[List[Identifiable], List[ReferenceInfo]]: A tuple containing the list of contained models and the list of references.
        """
        finder = cls()
        finder.model = model
        finder.find_contained_identifiables_and_references()
        return finder.contained_models, finder.references

    def find_contained_identifiables_and_references(self):
        """
        Method to find all contained identifiables (inclusive the model itself) and references in the model.
        """
        self.contained_models = get_all_contained_identifiables(self.model)
        self.references = get_reference_infos(self.contained_models)

    @classmethod
    def find_schema_references(
            cls,
            model: Identifiable,
    ) -> Tuple[List[Identifiable], Set[ReferenceInfo]]:
        """
        Method to find all contained models (inclusive the model itself) and references in a given model.

        Args:
            model (Identifiable): The model to find contained models and references in.

        Returns:
            Tuple[List[Identifiable], List[ReferenceInfo]]: A tuple containing the list of contained models and the list of references.
        """
        finder = cls()
        finder.model = model
        finder.find_contained_schemas_and_references()
        return finder.contained_schemas, finder.schema_references
    
    def find_contained_schemas_and_references(self):
        """
        Method to find all contained identifiables (inclusive the model itself) and references in the model.
        """
        self.contained_schemas = get_all_contained_schemas(self.model)
        self.schema_references = get_schema_reference_infos(self.contained_schemas)
        # FIXME: resolve empty referenced nodes without an associated type -> either find subclasses or classes that contain the referenced name (e.g. "acticePoleHousing" for class "PoleHousing")








from functools import partial
from typing import Tuple, Type

from aas_middleware.connect.connectors.connector import Connector
from aas_middleware.model.core import Identifiable


class PersistenceFactory:

    def __init__(self, connector_type: Type[Connector], *args, **kwargs):
        self.connector = partial(connector_type, *args, **kwargs)
    
    def create(self, model: Identifiable, *args, **kwargs) -> Connector:
        connector = self.connector(model, *args, **kwargs)
        return connector
        


from __future__ import annotations

from aas_middleware.connect.connectors.aas_client_connector.aas_client_connector import BasyxAASConnector, BasyxSubmodelConnector
from aas_middleware.middleware.middleware import Middleware
from aas_middleware.middleware.persistence_factory import PersistenceFactory
from aas_middleware.model.data_model import DataModel
from aas_middleware.model.data_model_rebuilder import DataModelRebuilder
from aas_middleware.model.formatting.aas.aas_model import AAS, Submodel

class AasMiddleware(Middleware):
    """
    Middleware that automatically that has aas and submodel repositories as persistence providers and consumers.
    """

    def __init__(self):
        super().__init__()

    def load_aas_persistent_data_model(self, name: str, data_model: DataModel, aas_host: str, aas_port: int, submodel_host: str, submodel_port: int, persist_instances: bool = False):
        """
        Function to load a data model into the middleware to be used for synchronization.

        Args:
            name (str): The name of the data model.
            data_model (DataModel): Data model containing the types and values.
            aas_host (str): The host of the AAS server.
            aas_port (int): The port of the AAS server.
            submodel_host (str): The host of the submodel server.
            submodel_port (int): The port of the submodel server.
            persist_instances (bool, optional): Whether to persist instances of the data model. Defaults to False.
        """
        # aas_data_model = DataModelRebuilder(data_model).rebuild_data_model_for_AAS_structure()
        aas_data_model = data_model
        self.load_data_model(name, aas_data_model, persist_instances)
        
        aas_persistence_factory = PersistenceFactory(BasyxAASConnector, host=aas_host, port=aas_port, submodel_host=submodel_host, submodel_port=submodel_port)
        submodel_persistence_factory = PersistenceFactory(BasyxSubmodelConnector, host=submodel_host, port=submodel_port)
        self.add_default_persistence(aas_persistence_factory, name, None, AAS)
        self.add_default_persistence(submodel_persistence_factory, name, None, Submodel)
        
    def scan_aas_server(self):
        """
        Function to scan the AAS server for all available AAS and Submodels.
        """
        # TODO: implement function
        pass


import asyncio
import functools
from inspect import signature
from types import NoneType
from typing import Dict, List
import typing

import anyio
from fastapi import APIRouter, BackgroundTasks, Body, HTTPException
from pydantic import BaseModel, Field, create_model
from aas_middleware.connect.workflows.worfklow_description import WorkflowDescription
from aas_middleware.connect.workflows.workflow import Workflow


def get_partial_type_hints(partial_func):
    original_func = partial_func.func
    original_hints = typing.get_type_hints(original_func)
    
    sig = signature(original_func)
    params = sig.parameters
    
    fixed_args = partial_func.args
    fixed_keywords = partial_func.keywords
    
    for index, param in enumerate(params.keys()):
        if index < len(fixed_args):
            original_hints.pop(param, None)
    
    for key in fixed_keywords.keys():
        original_hints.pop(key, None)
    
    return original_hints


def get_base_model_from_type_hints(name: str, type_hints: Dict[str, typing.Any]) -> typing.Type[BaseModel]:
    """
    Get the base model from the type hints of a function.

    Args:
        type_hints (Dict[str, typing.Any]): Type hints of a function.

    Returns:
        typing.Type[typing.Any]: Base model of the type hints.
    """
    if "return" in type_hints:
        type_hints.pop("return")
    dynamical_model_creation_dict = {}
    for argument, argument_type in type_hints.items():
        entry = {
            argument: typing.Annotated[
                argument_type, Field()
            ]
        }
        dynamical_model_creation_dict.update(entry)
    base_model = create_model(f"body_for_{name}", **dynamical_model_creation_dict)
    return base_model


def generate_workflow_endpoint(workflow: Workflow) -> List[APIRouter]:
    """
    Generates endpoints for a workflow to execute the workflow.

    Args:
        workflow (Workflow): Workflow that contains the function to be executed by the workflow.

    Returns:
        APIRouter: FastAPI router with an endpoint to execute the workflow.
    """
    router = APIRouter(
        prefix=f"/workflows/{workflow.get_name()}",
        tags=["workflows"],
        responses={404: {"description": "Not found"}},
    )

    if isinstance(workflow.workflow_function, functools.partial):
        type_hints = get_partial_type_hints(workflow.workflow_function)
    else:
        type_hints = typing.get_type_hints(workflow.workflow_function)
    
    return_type = type_hints.pop("return") if "return" in type_hints else None
    if len(type_hints) == 0:
        input_type_hints = None
    elif len(type_hints) == 1:
        input_type_hints = list(type_hints.values())[0]
    else:
        input_type_hints = get_base_model_from_type_hints(workflow.get_name(), type_hints)


    if input_type_hints is None:
        if workflow.get_description().interval is None:
            @router.post("/execute", response_model=return_type)
            async def execute():
                if workflow.running:
                    raise HTTPException(
                        status_code=400,
                        detail=f"Workflow {workflow.get_name()} is already running. Wait for it to finish or interrupt it first.",
                    )
                return await workflow.execute()

        @router.post("/execute_background", response_model=Dict[str, str])
        async def execute_background(background_tasks: BackgroundTasks):
            if workflow.running:
                raise HTTPException(
                    status_code=400,
                    detail=f"Workflow {workflow.get_name()} is already running. Wait for it to finish or interrupt it first.",
                )
            background_tasks.add_task(workflow.execute)
            return {"message": f"Started exeuction of workflow {workflow.get_name()}"}
    else:
        if workflow.get_description().interval is None:
            @router.post("/execute", response_model=return_type)
            # TODO: make the optional not here, but add a method that updates the POST endpoints after a connection is added, to have optional parameters...
            async def execute(arg: typing.Optional[input_type_hints]=None): # type: ignore
                if workflow.running:
                    raise HTTPException(
                        status_code=400,
                        detail=f"Workflow {workflow.get_name()} is already running. Wait for it to finish or interrupt it first.",
                    )
                if isinstance(arg, BaseModel) and len(type_hints) > 1:
                    input_value = dict(arg)
                    return await workflow.execute(**input_value)
                else:
                    return await workflow.execute(arg)

        @router.post("/execute_background", response_model=Dict[str, str])
        async def execute_background(background_tasks: BackgroundTasks, arg: typing.Optional[input_type_hints]=None): # type: ignore
            if workflow.running:
                raise HTTPException(
                    status_code=400,
                    detail=f"Workflow {workflow.get_name()} is already running. Wait for it to finish or interrupt it first.",
                )
            if isinstance(arg, BaseModel) and len(type_hints) > 1:
                input_value = dict(arg)
                background_tasks.add_task(workflow.execute, **input_value)
            else:
                background_tasks.add_task(workflow.execute, arg.model_dump())
            return {"message": f"Started exeuction of workflow {workflow.get_name()}"}
        

    @router.get("/description", response_model=WorkflowDescription)
    async def describe_workflow():
        return workflow.get_description()

    @router.get("/interrupt", response_model=Dict[str, str])
    async def interrupt_workflow():
        try:
            await workflow.interrupt()
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))
        return {"message": f"Stopped execution of workflow {workflow.get_name()}"}

    return router



from pyexpat import model
from typing import Generic, Protocol, TypeVar
import typing

from pydantic import BaseModel

from aas_middleware.model.data_model import DataModel

T = TypeVar("T", bound=BaseModel)
S = TypeVar("S", bound=BaseModel)


@typing.runtime_checkable
class Mapper(Protocol, Generic[S, T]):
    """
    Protocol for all mappers that are used to map data models to other data models.
    """

    def map(self, data: S) -> T:
        """
        Map a DataModel object to another DataModel object with different types and structure.

        Args:
            data (DataModel): A data model

        Returns:
            DataModel: The mapped data model.
        """
        ...

if __name__ == "__main__":
    class ExampleDataModel1(DataModel):
        pass

    class ExampleDataModel2(DataModel):
        pass


    class ModelMapper(Mapper[ExampleDataModel1, ExampleDataModel2]):
        def map(self, data: ExampleDataModel1) -> ExampleDataModel2:
            return ExampleDataModel2()

    example_data_model1 = ExampleDataModel1()
    example_data_model2 = ExampleDataModel2()

    model_mapper = ModelMapper()
    mapped_model = model_mapper.map(example_data_model1)
    print(mapped_model, type(mapped_model))  # <aas_middleware.model.mapping.mapper.ExampleDataModel2 object at 0x7f8b3c3b3d30> <class 'aas_middleware.model.mapping.mapper.ExampleDataModel2'>


    print(isinstance(model_mapper, Mapper))  # True
    print(typing.get_type_hints(model_mapper.map))
    print(typing.get_type_hints(ModelMapper.map))


import typing

from pydantic import BaseModel

from graphene_pydantic import PydanticObjectType
from graphene_pydantic.registry import get_global_registry

import graphene
from aas_middleware.connect.connectors.connector import Connector

if typing.TYPE_CHECKING:
    from aas_middleware.middleware.middleware import Middleware
from aas_middleware.middleware.registries import ConnectionInfo
from aas_middleware.model.data_model import DataModel
from starlette_graphene3 import (
    GraphQLApp,
    make_graphiql_handler,
)

from aas_middleware.model.formatting.aas.aas_middleware_util import get_all_submodel_elements_from_submodel, get_contained_models_attribute_info, union_type_check
from aas_middleware.model.formatting.aas.aas_model import AAS, Submodel, SubmodelElementCollection

def get_base_query_and_mutation_classes() -> (
    typing.Tuple[graphene.ObjectType, graphene.ObjectType]
):
    """
    Returns the base query and mutation classes for the GraphQL endpoint.

    Returns:
        tuple: Tuple of the base query and mutation classes for the GraphQL endpoint.
    """

    class Query(graphene.ObjectType):
        pass

    class Mutation(graphene.ObjectType):
        pass

    return Query, Mutation

class GraphQLRouter:
    def  __init__(self, data_model: DataModel, data_model_name: str, middleware: "Middleware"):
        self.data_model = data_model
        self.data_model_name = data_model_name

        self.middleware = middleware
        self.query, self.mutation = get_base_query_and_mutation_classes()

    def get_connector(self, item_id: str) -> Connector:
        return self.middleware.persistence_registry.get_connection(ConnectionInfo(data_model_name=self.data_model_name, model_id=item_id))
    

    def generate_graphql_endpoint(self):
        """
        Generates a GraphQL endpoint for the given data model and adds it to the middleware.
        """
        for top_level_model_type in self.data_model.get_top_level_types():
            self.create_query_for_model(top_level_model_type)
            # TODO: also make mutation possible
            # self.create_mutation_for_model(top_level_model_type)
        schema = graphene.Schema(query=self.query)
        graphql_app = GraphQLApp(schema=schema, on_get=make_graphiql_handler())
        self.middleware.app.mount("/graphql", graphql_app)

    def create_query_for_model(self, model_type: type):
        model_name = model_type.__name__

        submodels = get_contained_models_attribute_info(model_type)
        graphene_submodels = []
        for attribute_name, submodel in submodels:
            graphene_submodels.append(
                create_graphe_pydantic_output_type_for_submodel_elements(submodel)
            )

        for (attribute_name, submodel), graphene_submodel in zip(submodels, graphene_submodels):
            # FIXME: resolve problems with optional submodels!
            submodel_name = submodel.__name__
            class_dict = {
                f"{submodel_name}": graphene.List(graphene_submodel),
                f"resolve_{submodel_name}": self.get_submodel_resolve_function(submodel),
            }
            self.query = type("Query", (self.query,), class_dict)


        graphene_model = create_graphe_pydantic_output_type_for_model(model_type)

        class_dict = {
            f"{model_name}": graphene.List(graphene_model),
            f"resolve_{model_name}": self.get_aas_resolve_function(model_type),
        }
        self.query = type("Query", (self.query,), class_dict)

    def get_aas_resolve_function(self, model: typing.Type[BaseModel]) -> typing.Callable:
        """
        Returns the resolve function for the given pydantic model.

        Args:
            model (Type[BaseModel]): Pydantic model for which the resolve function should be created.

        Returns:
            typing.Callable: Resolve function for the given pydantic model.
        """
        middleware_instance = self.middleware
        async def resolve_models(self, info):
            aas_list = []
            connection_infos = middleware_instance.persistence_registry.get_type_connection_info(model.__name__)
            for connection_info in connection_infos:
                connector = middleware_instance.persistence_registry.get_connection(connection_info)
                retrieved_aas: AAS = await connector.provide()
                aas = model.model_validate(retrieved_aas.model_dump())
                aas_list.append(aas)
            return aas_list

        resolve_models.__name__ = f"resolve_{model.__name__}"
        return resolve_models


    def get_submodel_resolve_function(self, model: typing.Type[BaseModel]) -> typing.Callable:
        """
        Returns the resolve function for the given pydantic model.

        Args:
            model (Type[BaseModel]): Pydantic model for which the resolve function should be created.

        Returns:
            typing.Callable: Resolve function for the given pydantic model.
        """
        middleware_instance = self.middleware

        async def resolve_models(self, info):
            submodel_list = []
            # TODO: get the correct connectors here... no submodel connectors are available
            connection_infos = middleware_instance.persistence_registry.get_type_connection_info(model.__name__)
            for connection_info in connection_infos:
                connector = middleware_instance.persistence_registry.get_connection(connection_info)
                retrieved_submodel: Submodel = await connector.provide()
                submodel = model.model_validate(retrieved_submodel.model_dump())
                submodel_list.append(submodel)
            return submodel_list

        resolve_models.__name__ = f"resolve_{model.__name__}"
        return resolve_models


def add_class_method(model: typing.Type):
    def is_type_of(cls, root, info):
        return isinstance(root, (cls, model))

    class_method = classmethod(is_type_of)
    model.is_type_of = class_method


model_name_registry = set()


def create_graphe_pydantic_output_type_for_model(
    input_model: typing.Type[BaseModel], union_type: bool = False
) -> PydanticObjectType:
    """
    Creates a pydantic model for the given pydantic model.

    Args:
        model (Type[BaseModel]): Pydantic model for which the Graphene Object Type should be created.

    Returns:
        PydanticObjectType: Graphene Object type for the given pydantic model.
    """
    graphene_model_registry = get_global_registry(PydanticObjectType)._registry
    for model in graphene_model_registry.keys():
        if input_model == model.__name__:
            return graphene_model_registry[model]

    rework_default_list_to_default_factory(input_model)
    graphene_model = type(
        input_model.__name__,
        (PydanticObjectType,),
        {"Meta": type("Meta", (), {"model": input_model})},
    )
    if union_type:
        add_class_method(graphene_model)

    return graphene_model


def is_typing_list_or_tuple(input_type: typing.Any) -> bool:
    """
    Checks if the given type is a typing.List or typing.Tuple.

    Args:
        input_type (typing.Any): Type to check.

    Returns:
        bool: True if the given type is a typing.List or typing.Tuple, False otherwise.
    """
    return typing.get_origin(input_type) == list or typing.get_origin(input_type) == tuple


def list_contains_any_submodel_element_collections(
    input_type: typing.Union[typing.List, typing.Tuple]
) -> bool:
    return any(
        issubclass(nested_type, SubmodelElementCollection)
        for nested_type in typing.get_args(input_type)
    )


def rework_default_list_to_default_factory(model: BaseModel):
    for names, field in model.model_fields.items():
        if field.default:
            pass
        if (
            isinstance(field.default, list)
            or isinstance(field.default, tuple)
            or isinstance(field.default, set)
        ):
            if field.default:
                field.annotation = type(field.default[0])
            else:
                # TODO: potentially remove this...
                field.annotation = typing.List[str]
            field.default = None
        if isinstance(field.default, BaseModel):
            field.default = None


def create_graphe_pydantic_output_type_for_submodel_elements(
    model: Submodel, union_type: bool = False
) -> PydanticObjectType:
    """
    Create recursively graphene pydantic output types for submodels and submodel elements.

    Args:
        model (typing.Union[base.Submodel, base.SubmodelElementCollectiontuple, list, set, ]): Submodel element for which the graphene pydantic output types should be created.
    """
    for attribute_name, attribute_value in get_all_submodel_elements_from_submodel(
        model
    ).items():
        if union_type_check(attribute_value):
            subtypes = typing.get_args(attribute_value)
            for subtype in subtypes:
                create_graphe_pydantic_output_type_for_submodel_elements(
                    subtype, union_type=True
                )
        elif hasattr(attribute_value, "model_fields") and issubclass(
            attribute_value, SubmodelElementCollection
        ):
            create_graphe_pydantic_output_type_for_submodel_elements(attribute_value)
        elif is_typing_list_or_tuple(attribute_value):
            if not list_contains_any_submodel_element_collections(attribute_value):
                continue
            for nested_type in typing.get_args(attribute_value):
                if union_type_check(nested_type):
                    subtypes = typing.get_args(nested_type)
                    for subtype in subtypes:
                        create_graphe_pydantic_output_type_for_submodel_elements(
                            subtype, union_type=True
                        )
                elif issubclass(nested_type, SubmodelElementCollection):
                    create_graphe_pydantic_output_type_for_submodel_elements(
                        nested_type
                    )
    return create_graphe_pydantic_output_type_for_model(model, union_type)